#------------------------------------------------------------------------------#

# Import pyCAPS module
import pyCAPS

# Import os
import os

#------------------------------------------------------------------------------#
# Define units
m    = pyCAPS.Unit("meter")
kg   = pyCAPS.Unit("kg")
s    = pyCAPS.Unit("s")
K    = pyCAPS.Unit("Kelvin")
deg  = pyCAPS.Unit("degree")
kft  = pyCAPS.Unit("kft")

unitSystem={"mass":kg, "length":m, "time":s, "temperature":K}

#------------------------------------------------------------------------------#

# Load geometry [.csm] file
filename = os.path.join("..","ESP","transport.csm")
print ("\n==> Loading geometry from file \""+filename+"\"...")
myProblem = pyCAPS.Problem(problemName = "workDir_5_Link",
                           capsFile = filename,
                           outLevel = 1)

# Alias the geometry
transport = myProblem.geometry

# Change to VLM view
transport.cfgpmtr.VIEW.Concept = 0
transport.cfgpmtr.VIEW.VLM     = 1

# Enable lifting surfaces without controls
transport.cfgpmtr.COMP.Wing    = 1
transport.cfgpmtr.COMP.Fuse    = 0
transport.cfgpmtr.COMP.Htail   = 1
transport.cfgpmtr.COMP.Vtail   = 1
transport.cfgpmtr.COMP.Control = 0

# Create shared analysis input parameter
mach = myProblem.parameter.create("Mach", 0.1)

# Create friction AIM
friction = myProblem.analysis.create(aim = "frictionAIM",
                                     name = "friction")

# Link mach number and set altitude
friction.input["Mach"].link(mach)
friction.input.Altitude = 5*kft

# Create AVL AIM
avl = myProblem.analysis.create(aim = "avlAIM",
                                name = "avl",
                                unitSystem = unitSystem)

# Link inputs
avl.input["Mach"].link(mach)
avl.input["CDp"].link(friction.output["CDtotal"])

# Set analysis specific variables
avl.input.Alpha = 1.5*deg
avl.input.Beta  = 0.0*deg

# Set meshing parameters for each surface
wing = {"numChord"     : 4,
        "numSpanTotal" : 80}

htail = {"numChord"     : 4,
         "numSpanTotal" : 30}

vtail = {"numChord"     : 4,
         "numSpanTotal" : 10}

# Associate the surface parameters with capsGroups defined on the geometry
avl.input.AVL_Surface = {"Wing" : wing ,
                         "Htail": htail,
                         "Vtail": vtail}

# Compute drag for a range of Mach numbers
CDtotal = []
CDp = []
machRange = [mach.value + i*0.1 for i in range(7)]
for M in machRange:
    # Set mission parameters - AIMs will be updated automatically due to link
    mach.value = M
    
    # Run AIM pre-analysis
    print ("\n==> Running friction preAnalysis")
    friction.preAnalysis()
    
    ####### Run friction ####################
    print ("\n\n==> Running friction......")
    
    currentDirectory = os.getcwd() # Get current working directory
    os.chdir(friction.analysisDir) # Move into test directory
    
    # Run avl via system call
    os.system("friction frictionInput.txt frictionOutput.txt > Info.out");
    
    os.chdir(currentDirectory)     # Move back to top directory
    ######################################
    
    # Run AIM post-analysis
    print ("\n==> Running friction postAnalysis")
    friction.postAnalysis()

    # Track profile drag added to AVL
    CDp.append(friction.output.CDtotal)

    # Run AIM pre-analysis
    print ("\n==> Running avl preAnalysis")
    avl.preAnalysis()
    
    ####### Run avl ####################
    print ("\n\n==> Running avl......")
    
    currentDirectory = os.getcwd() # Get current working directory
    os.chdir(avl.analysisDir)      # Move into test directory
    
    # Run avl via system call
    os.system("avl caps < avlInput.txt > avlOutput.txt");
    
    os.chdir(currentDirectory)     # Move back to top directory
    ######################################
    
    # Run AIM post-analysis
    print ("\n==> Running avl postAnalysis")
    avl.postAnalysis()
    
    # Track the total drag
    CDtotal.append(avl.output.CDtot)

print()
print("--> Mach =", machRange)
print("--> CDp  =", CDp)
print("--> CD   =", CDtotal)
print()
