#------------------------------------------------------------------------------#

# Import pyCAPS module
import pyCAPS

# Import os module
import os
import shutil

#------------------------------------------------------------------------------#

# Load geometry [.csm] file
filename = os.path.join("..","ESP","wing3.csm")
print ("\n==> Loading geometry from file \""+filename+"\"...")
myProblem = pyCAPS.Problem(problemName = "workDir_1_ModalClamped",
                           capsFile = filename,
                           outLevel = 1)

# Alias the geometry
wing = myProblem.geometry

# Set geometry variables to enable Box Structure with Clambed root
wing.cfgpmtr.VIEW.Concept          = 0
wing.cfgpmtr.VIEW.ClampedStructure = 1
wing.cfgpmtr.VIEW.BoxStructure     = 1

#------------------------------------------------------------------------------#

# Create EGADS tess aim
tess = myProblem.analysis.create(aim = "egadsTessAIM",
                                 name = "tess")

# No Tess vertexes on edges (minimial mesh)
tess.input.Edge_Point_Max = 2

# Use regularized quads
tess.input.Mesh_Elements = "Quad"

# Run AIM pre/post-analysis to generate the surface mesh
print ("\n==> Generating Mesh...")
tess.preAnalysis()
tess.postAnalysis()

#------------------------------------------------------------------------------#

# Create astros aim
astros = myProblem.analysis.create(aim = "astrosAIM",
                                   name = "astros")

# Link the EGADS tess Surface_Mesh to astros
astros.input["Mesh"].link(tess.output["Surface_Mesh"])

# Set analysis specific variables
astros.input.Proj_Name        = "astros_modal"
astros.input.Mesh_File_Format = "Large"

# Set constraints
constraint = {"dofConstraint" : 123456}
astros.input.Constraint = {"rootConstraint": constraint}

# Set analysis type
eigen = { "analysisType"         : "Modal",
          "extractionMethod"     : "SINV",
          "frequencyRange"       : [0, 10],
          "numEstEigenvalue"     : 1,
          "numDesiredEigenvalue" : 10,
          "eigenNormaliztion"    : "MASS"}
astros.input.Analysis = {"EigenAnalysis": eigen}

# Set materials
unobtainium  = {"youngModulus" : 2.2E6 ,
                "poissonRatio" : .5,
                "density"      : 7850}
madeupium    = {"materialType" : "isotropic",
                "youngModulus" : 1.2E5 ,
                "poissonRatio" : .5,
                "density"      : 7850}
astros.input.Material = {"Unobtainium": unobtainium,
                         "Madeupium"  : madeupium}

# Set properties
skinShell  = {"propertyType"        : "Shell",
              "material"            : "unobtainium",
              "bendingInertiaRatio" : 1.0,
              "shearMembraneRatio"  : 0, # Turn of shear - no materialShear
              "membraneThickness"   : 0.05}
ribShell  = {"propertyType"        : "Shell",
             "material"            : "unobtainium",
             "bendingInertiaRatio" : 1.0,
             "shearMembraneRatio"  : 0, # Turn of shear - no materialShear
             "membraneThickness"   : 0.1}
sparShell  = {"propertyType"        : "Shell",
              "material"            : "madeupium",
              "bendingInertiaRatio" : 1.0,
              "shearMembraneRatio"  : 0, # Turn of shear - no materialShear
              "membraneThickness"   : 0.2}
astros.input.Property = {"leftWingSkin": skinShell,
                         "riteWingSkin": skinShell,
                         "wingRib"     :  ribShell,
                         "wingSpar1"   : sparShell,
                         "wingSpar2"   : sparShell}

# Run AIM pre-analysis
print ("\n==> Running ASTROS pre-analysis...")
astros.preAnalysis()

print ("\n==> Running mASTROS...")
####### Run ASTROS ####################
currentDirectory = os.getcwd()   # Get current working directory
os.chdir(astros.analysisDir)     # Move into test directory

# Declare ASTROS install directory
astrosInstallDir = os.environ['ESP_ROOT'] + os.sep + "bin" + os.sep

# Copy files needed to run astros
files = ["ASTRO.D01", "ASTRO.IDX"]
for file in files:
    try:
        shutil.copy2(astrosInstallDir + file, file)
    except:
        print ('Unable to copy "' + file + '"')
        raise

# Run micro-ASTROS via system call
os.system("mastros.exe < " + astros.input.Proj_Name +  ".dat > " + astros.input.Proj_Name + ".out")

# Remove temporary files
for file in files:
    if os.path.isfile(file):
        os.remove(file)

os.chdir(currentDirectory)       # Move back to top directory
#######################################

# Run AIM post-analysis
astros.postAnalysis()

# Get list of Eigen-frequencies
freqs = astros.output.EigenFrequency

print ("\n--> Eigen-frequencies:")
for i in range(len(freqs)):
    print ("    " + repr(i+1).ljust(2) + ": " + str(freqs[i]))
