#------------------------------------------------------------------------------#

# Import pyCAPS module
import pyCAPS

# Import os module
import os

#------------------------------------------------------------------------------#

# Load geometry [.csm] file
filename = os.path.join("..","ESP","wing3.csm")
print ("\n==> Loading geometry from file \""+filename+"\"...")
myProblem = pyCAPS.Problem(problemName = "workDir_5_Flutter",
                           capsFile = filename,
                           outLevel = 1)

# Alias the geometry
wing = myProblem.geometry

# Change to Structures and VLM
wing.cfgpmtr.VIEW.Concept          = 0
wing.cfgpmtr.VIEW.VLM              = 1
wing.cfgpmtr.VIEW.ClampedStructure = 1

#------------------------------------------------------------------------------#

# Create EGADS tess aim
tess = myProblem.analysis.create(aim = "egadsTessAIM",
                                 name = "tess")

# No Tess vertexes on edges (minimial mesh)
tess.input.Edge_Point_Max = 2

# Use regularized quads
tess.input.Mesh_Elements = "Quad"

# Run AIM pre/post-analysis to generate the surface mesh
print ("\n==> Generating Mesh...")
tess.preAnalysis()
tess.postAnalysis()

#------------------------------------------------------------------------------#

# Create nastran aim
nastran = myProblem.analysis.create(aim  = "nastranAIM",
                                    name = "nastran")

# Link the EGADS tess Surface_Mesh to nastran
nastran.input["Mesh"].link(tess.output["Surface_Mesh"])

# Set analysis specific variables
nastran.input.Proj_Name        = "nastran_modal"
nastran.input.Mesh_File_Format = "Large"

# Set constraints
constraint = {"dofConstraint" : 123456}
nastran.input.Constraint = {"rootConstraint": constraint}

# Set materials
unobtainium  = {"youngModulus" : 2.2e6 ,
                "poissonRatio" : .5,
                "density"      : 7850}
madeupium    = {"materialType" : "isotropic",
                "youngModulus" : 1.2e5 ,
                "poissonRatio" : .5,
                "density"      : 7850}
nastran.input.Material = {"Unobtainium": unobtainium,
                          "Madeupium"  : madeupium}

# Set properties
skinShell  = {"propertyType"        : "Shell",
              "material"            : "unobtainium",
              "bendingInertiaRatio" : 1.0,
              "shearMembraneRatio"  : 0, # Turn of shear - no materialShear
              "membraneThickness"   : 0.05}
ribShell  = {"propertyType"        : "Shell",
             "material"            : "unobtainium",
             "bendingInertiaRatio" : 1.0,
             "shearMembraneRatio"  : 0, # Turn of shear - no materialShear
             "membraneThickness"   : 0.1}
sparShell  = {"propertyType"        : "Shell",
              "material"            : "madeupium",
              "bendingInertiaRatio" : 1.0,
              "shearMembraneRatio"  : 0, # Turn of shear - no materialShear
              "membraneThickness"   : 0.2}
nastran.input.Property = {"leftWingSkin": skinShell,
                          "riteWingSkin": skinShell,
                          "wingRib"     :  ribShell,
                          "wingSpar1"   : sparShell,
                          "wingSpar2"   : sparShell}

# Set analysis type
nastran.input.Analysis_Type = "AeroelasticFlutter"

# Aero with capsGroup for airfoil sections
wingVLM = {"groupName"         : "Wing",
           "numChord"          : 4,
           "numSpanPerSection" : 6}

# Note the surface name corresponds to the capsBound found in the *.csm file. This links
# the spline for the aerodynamic surface to the structural model
nastran.input.VLM_Surface = {"upperWing": wingVLM}
    
# Set analysis 
flutter = { "analysisType"         : "AeroelasticFlutter",
            "extractionMethod"     : "Lanczos",
            "frequencyRange"       : [0.1, 300],
            "numEstEigenvalue"     : 4,
            "numDesiredEigenvalue" : 4, 
            "eigenNormaliztion"    : "MASS", 
            "aeroSymmetryXY"       : "ASYM",
            "aeroSymmetryXZ"       : "SYM",
            "analysisConstraint"   : ["rootConstraint"],
            "machNumber"           : 0.1029,
            "dynamicPressure"      : 0.5* 1.2e-6 * 3500**2,
            "density"              : 1.2e-6, 
            "reducedFreq"          : [0.001, 0.01, 0.1, 0.12, 0.14, 0.16, 0.18, 0.25],
            }
nastran.input.Analysis = {"Flutter": flutter}

# Run AIM pre-analysis
print ("\n==> Running Nastran pre-analysis...")
nastran.preAnalysis()

print ("\n==> Running Nastran...")
####### Run NASTRAN ###################
currentDirectory = os.getcwd()   # Get current working directory
os.chdir(nastran.analysisDir)    # Move into test directory

# Run nastran via system call
os.system("nastran old=no notify=no batch=no scr=yes sdirectory=./ " + nastran.input.Proj_Name +  ".dat"); # Run Nastran via system call

os.chdir(currentDirectory)       # Move back to original directory
#######################################

# Run AIM post-analysis
nastran.postAnalysis()

