#------------------------------------------------------------------------------#

# Import pyCAPS module
import pyCAPS

# Import os module
import os
import shutil

#------------------------------------------------------------------------------#

myProblem = pyCAPS.Problem(problemName = "workDir_1_Aeroelastic_Modal",
                           capsFile = os.path.join("..","ESP","wing3.csm"))

# Create an alias for geometry
wing = myProblem.geometry

# Enable structures and CFD views
wing.cfgpmtr.VIEW.Concept          = 0
wing.cfgpmtr.VIEW.ClampedStructure = 1
wing.cfgpmtr.VIEW.CFDInviscid      = 1

#------------------------------------------------------------------------------#

# Create projectName variable
projectName = "aeroelasticModal"

#------------------------------------------------------------------------------#

# Create aflr4 AIM
aflr4 = myProblem.analysis.create(aim = "aflr4AIM",
                                  name = "aflr4")

# Farfield growth factor
aflr4.input.ff_cdfr = 1.4

# Scaling factor to compute AFLR4 'ref_len' parameter
aflr4.input.Mesh_Length_Factor = 5

# Edge mesh spacing discontinuity scaled interpolant and farfield BC
aflr4.input.Mesh_Sizing = {"leftWing": {"edgeWeight":1.0},
                           "riteWing": {"edgeWeight":1.0},
                           "Farfield": {"bcType":"Farfield"}}

#------------------------------------------------------------------------------#

# Create AFLR3 AIM
aflr3 = myProblem.analysis.create(aim  = "aflr3AIM",
                                  name = "aflr3")

# Link the surface mesh
aflr3.input["Surface_Mesh"].link(aflr4.output["Surface_Mesh"])

#------------------------------------------------------------------------------#

# Create fun3d AIM
fun3d = myProblem.analysis.create(aim  = "fun3dAIM",
                                  name = "fun3d")

# Link the volume mesh
fun3d.input["Mesh"].link(aflr3.output["Volume_Mesh"])

# Set inputs for fun3d
speedofSound = 340.0 # m/s
refVelocity = 100.0 # m/s
refDensity = 1.2 # kg/m^3

fun3d.input.Proj_Name       = projectName
fun3d.input.Mesh_ASCII_Flag = False
fun3d.input.Mach            = refVelocity/speedofSound
fun3d.input.Equation_Type   = "compressible"
fun3d.input.Viscous         = "inviscid"
fun3d.input.Num_Iter        = 10
fun3d.input.Time_Accuracy   = "2ndorderOPT"
fun3d.input.Time_Step       = 0.001*speedofSound
fun3d.input.Num_Subiter     = 30
fun3d.input.Temporal_Error  = 0.01

fun3d.input.CFL_Schedule      = [1, 5.0]
fun3d.input.CFL_Schedule_Iter = [1, 40]
fun3d.input.Overwrite_NML     = True
fun3d.input.Restart_Read      = "off"

inviscidBC = {"bcType" : "Inviscid"}
fun3d.input.Boundary_Condition = {"Wing": inviscidBC,
                                  "Farfield": "farfield"}

#------------------------------------------------------------------------------#

# Create EGADS tess aim
tess = myProblem.analysis.create(aim = "egadsTessAIM")

# No Tess vertexes on edges (minimial mesh)
tess.input.Edge_Point_Max = 2

# Use regularized quads
tess.input.Mesh_Elements = "Quad"

#------------------------------------------------------------------------------#

# Create ASTROS AIMs
astros = myProblem.analysis.create(aim  = "astrosAIM",
                                   name = "astros")

# Set inputs for astros
astros.input.Proj_Name = projectName

astros.input.Analysis_Type = "Modal"

# Set analysis type
eigen = { "analysisType"         : "Modal",
          "extractionMethod"     : "SINV",
          "frequencyRange"       : [0, 10],
          "numEstEigenvalue"     : 1,
          "numDesiredEigenvalue" : 10,
          "eigenNormaliztion"    : "MASS"}
astros.input.Analysis = {"EigenAnalysis": eigen}

# Set constraints
constraint = {"groupName"     : ["rootConstraint"],
              "dofConstraint" : 123456}
astros.input.Constraint = {"rootConstraint": constraint}

# Set materials
unobtainium  = {"youngModulus" : 2.2E6 ,
                "poissonRatio" : .5,
                "density"      : 7850}
madeupium    = {"materialType" : "isotropic",
                "youngModulus" : 1.2E5 ,
                "poissonRatio" : .5,
                "density"      : 7850}
astros.input.Material = {"Unobtainium": unobtainium,
                         "Madeupium"  : madeupium}

# Set properties
skinShell  = {"propertyType"        : "Shell",
              "material"            : "unobtainium",
              "bendingInertiaRatio" : 1.0,
              "shearMembraneRatio"  : 0, # Turn of shear - no materialShear
              "membraneThickness"   : 0.05}
ribShell  = {"propertyType"        : "Shell",
             "material"            : "unobtainium",
             "bendingInertiaRatio" : 1.0,
             "shearMembraneRatio"  : 0, # Turn of shear - no materialShear
             "membraneThickness"   : 0.1}
sparShell  = {"propertyType"        : "Shell",
              "material"            : "madeupium",
              "bendingInertiaRatio" : 1.0,
              "shearMembraneRatio"  : 0, # Turn of shear - no materialShear
              "membraneThickness"   : 0.2}
astros.input.Property = {"leftWingSkin": skinShell,
                         "riteWingSkin": skinShell,
                         "wingRib"     :  ribShell,
                         "wingSpar1"   : sparShell,
                         "wingSpar2"   : sparShell}

#------------------------------------------------------------------------------#

# Create an array of EigenVector names
numEigenVector = 3
eigenVectors = []
for i in range(numEigenVector):
    eigenVectors.append("EigenVector_" + str(i+1))

# Create the capsBounds for data transfer
boundNames = ["upperWing", "lowerWing", "leftTip", "riteTip"]
for boundName in boundNames:
    # Create the bound
    bound = myProblem.bound.create(boundName)
    
    # Create the vertex sets on the bound for fun3d and astros analysis
    fun3dVset  = bound.vertexSet.create(fun3d)
    astrosVset = bound.vertexSet.create(astros)
    
    # Create eigenVector data sets
    for eigenVector in eigenVectors:
        fun3d_eigenVector  = fun3dVset.dataSet.create(eigenVector, pyCAPS.fType.FieldIn)
        astros_eigenVector = astrosVset.dataSet.create(eigenVector, pyCAPS.fType.FieldOut)

        # Link the data sets
        fun3d_eigenVector.link(astros_eigenVector, "Interpolate")
    
    # Close the bound as complete (cannot create more vertex or data sets)
    bound.close()

#------------------------------------------------------------------------------#

# Run AIM pre/post-analysis to generate the meshes
for aim in [aflr4.name, aflr3.name, tess.name]:
    myProblem.analysis[aim].preAnalysis()
    myProblem.analysis[aim].postAnalysis()

#------------------------------------------------------------------------------#

# Run AIM pre-analysis
print ("\n==> Running ASTROS pre-analysis...")
astros.preAnalysis()

####### Run ASTROS ####################
print ("\n==> Running mASTROS...")

# Get our current working directory (so that we can return here after running ASTROS)
currentPath = os.getcwd()

# Move into temp directory
os.chdir(astros.analysisDir)

# Declare ASTROS install directory
astrosInstallDir = os.environ['ESP_ROOT'] + os.sep + "bin" + os.sep

# Copy files needed to run astros
files = ["ASTRO.D01", "ASTRO.IDX"]
for file in files:
    try:
        shutil.copy2(astrosInstallDir + file, file)
    except:
        print ("Unable to copy \"", file, "\"")
        raise SystemError

# Run micro-ASTROS via system call
os.system("mastros.exe < " + projectName +  ".dat > " + projectName + ".out")

# Remove temporary files
for file in files:
    if os.path.isfile(file):
        os.remove(file)

# Move back to original directory
os.chdir(currentPath)
#######################################

# Run AIM post-analysis
astros.postAnalysis()

#------------------------------------------------------------------------------#

# Retrive natural frequencies from the structural analysis
naturalFreq = astros.output.EigenRadian # rads/s
mass        = astros.output.EigenGeneralMass

modalDict = {} # Build up Modal Aeroelastic dict
for j in eigenVectors:
    modeNum = int(j.strip("EigenVector_"))

    value = {"frequency" : naturalFreq[modeNum-1],
             "damping" : 0.99,
             "generalMass" : mass[modeNum-1],
             "generalVelocity" : 0.1}

    modalDict[j] = value

# Add Eigen information in fun3d
fun3d.input.Modal_Aeroelastic          = modalDict
fun3d.input.Modal_Ref_Velocity         = refVelocity
fun3d.input.Modal_Ref_Dynamic_Pressure = 0.5*refDensity*refVelocity*refVelocity
fun3d.input.Modal_Ref_Length           = 1.0

# Run the preAnalysis
print ("\n==> Running FUN3D pre-analysis...")
fun3d.preAnalysis()

####### Run fun3d ####################
print ("\n\nRunning FUN3D......")
currentDirectory = os.getcwd() # Get our current working directory

os.chdir(fun3d.analysisDir) # Move into test directory

cmdLineOpt = "--moving_grid --aeroelastic_internal --animation_freq -1"

# Run fun3d via system call
os.system("mpirun -np 4 nodet_mpi " + cmdLineOpt + " > Info.out");

os.chdir(currentDirectory) # Move back to top directory
#######################################

# Run AIM post-analysis
fun3d.postAnalysis()
