#------------------------------------------------------------------------------#

# Import pyCAPS module
import pyCAPS

# Import os module
import os
import shutil

# Import SU2 python environment
from parallel_computation import parallel_computation as su2Run
from mesh_deformation import mesh_deformation as su2Deform

#------------------------------------------------------------------------------#

# Load CSM file
myProblem = pyCAPS.Problem(problemName = "workDir_2_Aeroelastic_Iterative",
                           capsFile = os.path.join("..","ESP","wing3.csm"))

# Create an alias for geometry
wing = myProblem.geometry

# Enable structures and CFD views
wing.cfgpmtr.VIEW.Concept          = 0
wing.cfgpmtr.VIEW.ClampedStructure = 1
wing.cfgpmtr.VIEW.CFDInviscid      = 1

#------------------------------------------------------------------------------#

# Create projectName vairbale
projectName = "aeroelasticIterative"

# Set the number of transfer iterations
numTransferIteration = 2

#------------------------------------------------------------------------------#

# Create aflr4 AIM
aflr4 = myProblem.analysis.create(aim = "aflr4AIM",
                                  name = "aflr4")

# Farfield growth factor
aflr4.input.ff_cdfr = 1.4

# Scaling factor to compute AFLR4 'ref_len' parameter
aflr4.input.Mesh_Length_Factor = 5

# Edge mesh spacing discontinuity scaled interpolant and farfield BC
aflr4.input.Mesh_Sizing = {"leftWing": {"edgeWeight":1.0},
                           "riteWing": {"edgeWeight":1.0},
                           "Farfield": {"bcType":"Farfield"}}

#------------------------------------------------------------------------------#

# Create AFLR3 AIM
aflr3 = myProblem.analysis.create(aim  = "aflr3AIM",
                                  name = "aflr3")

# Link the surface mesh
aflr3.input["Surface_Mesh"].link(aflr4.output["Surface_Mesh"])

#------------------------------------------------------------------------------#

# Create SU2 AIMs
su2 = myProblem.analysis.create(aim = "su2AIM",
                                name  = "su2")

# Link the volume mesh
su2.input["Mesh"].link(aflr3.output["Volume_Mesh"])

# Set inputs for su2
speedofSound = 340.0 # m/s
refVelocity = 100.0 # m/s
refDensity = 1.2 # kg/m**3

su2.input.Proj_Name     = projectName
su2.input.SU2_Version   = "Blackbird"
su2.input.Mach          = refVelocity/speedofSound
su2.input.Equation_Type = "compressible"
su2.input.Num_Iter      = 5 # Way too few to converge the solver, but this is an example
su2.input.Output_Format = "Tecplot"
su2.input.Overwrite_CFG = True
su2.input.Pressure_Scale_Factor = 0.5*refDensity*refVelocity**2
su2.input.Surface_Monitor = "Wing"

inviscidBC = {"bcType" : "Inviscid"}
su2.input.Boundary_Condition = {"Wing": inviscidBC,
                                "Farfield": "farfield"}

#------------------------------------------------------------------------------#

# Create EGADS tess aim
tess = myProblem.analysis.create(aim = "egadsTessAIM")

# No Tess vertexes on edges (minimial mesh)
tess.input.Edge_Point_Max = 2

# Use regularized quads
tess.input.Mesh_Elements = "Quad"

#------------------------------------------------------------------------------#

# Create ASTROS AIMs
astros = myProblem.analysis.create(aim  = "astrosAIM",
                                   name = "astros")

# Link the surface mesh
astros.input["Mesh"].link(tess.output["Surface_Mesh"])

# Set inputs for astros
astros.input.Proj_Name = projectName

# Set analysis type
astros.input.Analysis_Type = "Static"

# External pressure load to astros that we will inherited from su2
load = {"loadType" : "PressureExternal"}
astros.input.Load = {"pressureAero": load}

# Set constraints
constraint = {"groupName"     : ["rootConstraint"],
              "dofConstraint" : 123456}
astros.input.Constraint = {"rootConstraint": constraint}

# Set materials
unobtainium  = {"youngModulus" : 2.2e6,
                "poissonRatio" : .5,
                "density"      : 7850}
madeupium    = {"materialType" : "isotropic",
                "youngModulus" : 1.2e5,
                "poissonRatio" : .5,
                "density"      : 7850}
astros.input.Material = {"Unobtainium": unobtainium,
                         "Madeupium"  : madeupium}

# Set properties
skinShell  = {"propertyType"        : "Shell",
              "material"            : "unobtainium",
              "bendingInertiaRatio" : 1.0,
              "shearMembraneRatio"  : 0, # Turn of shear - no materialShear
              "membraneThickness"   : 0.05}
ribShell  = {"propertyType"        : "Shell",
             "material"            : "unobtainium",
             "bendingInertiaRatio" : 1.0,
             "shearMembraneRatio"  : 0, # Turn of shear - no materialShear
             "membraneThickness"   : 0.1}
sparShell  = {"propertyType"        : "Shell",
              "material"            : "madeupium",
              "bendingInertiaRatio" : 1.0,
              "shearMembraneRatio"  : 0, # Turn of shear - no materialShear
              "membraneThickness"   : 0.2}
astros.input.Property = {"leftWingSkin": skinShell,
                         "riteWingSkin": skinShell,
                         "wingRib"     :  ribShell,
                         "wingSpar1"   : sparShell,
                         "wingSpar2"   : sparShell}

#------------------------------------------------------------------------------#

# Create the data transfer connections
boundNames = ["upperWing", "lowerWing", "leftTip", "riteTip"]
for boundName in boundNames:
    # Create the bound
    bound = myProblem.bound.create(boundName)
    
    # Create the vertex sets on the bound for su2 and astros analysis
    su2Vset    = bound.vertexSet.create(su2)
    astrosVset = bound.vertexSet.create(astros)
    
    # Create pressure data sets
    su2_Pressure    = su2Vset.dataSet.create("Pressure", pyCAPS.fType.FieldOut)
    astros_Pressure = astrosVset.dataSet.create("Pressure", pyCAPS.fType.FieldIn)

    # Create displacement data sets
    su2_Displacement    = su2Vset.dataSet.create("Displacement", pyCAPS.fType.FieldIn, init=[0,0,0])
    astros_Displacement = astrosVset.dataSet.create("Displacement", pyCAPS.fType.FieldOut)

    # Link the data sets
    astros_Pressure.link(su2_Pressure, "Conserve")
    su2_Displacement.link(astros_Displacement, "Interpolate")
    
    # Close the bound as complete (cannot create more vertex or data sets)
    bound.close()

#------------------------------------------------------------------------------#

# Run AIM pre/post-analysis to generate the meshes
for aim in [aflr4.name, aflr3.name, tess.name]:
    myProblem.analysis[aim].preAnalysis()
    myProblem.analysis[aim].postAnalysis()

#------------------------------------------------------------------------------#

# Declare ASTROS install directory
astrosInstallDir = os.environ['ESP_ROOT'] + os.sep + "bin" + os.sep

# Copy files needed to run astros
files = ["ASTRO.D01", "ASTRO.IDX"]
for file in files:
    try:
        shutil.copy2(astrosInstallDir + file, astros.analysisDir + os.sep + file)
    except:
        print ("Unable to copy \"", file, "\"")
        raise SystemError

# Aeroelastic iteration loop
for iter in range(numTransferIteration):

    ####### SU2 ###########################
    print ("\n==> Running SU2 pre-analysis...")
    su2.preAnalysis()

    #------------------------------
    print ("\n\nRunning SU2......")
    currentDirectory = os.getcwd() # Get our current working directory

    os.chdir(su2.analysisDir) # Move into test directory

    numberProc = 1
    # Run SU2 mesh deformation
    if iter > 0:
        su2Deform(projectName + ".cfg", numberProc)

    # Run SU2 solver
    su2Run(projectName + ".cfg", numberProc)

    shutil.copy("surface_flow_" + projectName + ".szplt", "surface_flow_" + projectName + "_Iteration_" + str(iter) + ".szplt")
    os.chdir(currentDirectory) # Move back to top directory
    #------------------------------

    print ("\n==> Running SU2 post-analysis...")
    su2.postAnalysis()
    #######################################


    ####### ASTROS ########################
    print ("\n==> Running ASTROS pre-analysis...")
    astros.preAnalysis()

    #------------------------------
    print ("\n\nRunning mASTROS......")
    currentDirectory = os.getcwd() # Get our current working directory

    os.chdir(astros.analysisDir) # Move into test directory

    # Run mASTROS via system call
    os.system("mastros.exe < " + projectName +  ".dat > " + projectName + ".out");

    os.chdir(currentDirectory) # Move back to top directory
    #------------------------------

    print ("\n==> Running ASTROS post-analysis...")
    astros.postAnalysis()
    #######################################
